/* foldDb - Search database info for making foldout. */
#include "common.h"
#include "hash.h"
#include "linefile.h"
#include "obscure.h"
#include "jksql.h"
#include "hdb.h"
#include "memgfx.h"
#include "psl.h"
#include "genePred.h"
#include "est3.h"
#include "gcPercent.h"
#include "rmskOut.h"
#include "agpFrag.h"
#include "agpGap.h"
#include "cpgIsland.h"
#include "wabAli.h"
#include "clonePos.h"
#include "ctgPos.h"
#include "glDbRep.h"
#include "knownInfo.h"
#include "exoFish.h"
#include "rnaGene.h"
#include "snp.h"
#include "cytoBand.h"

/* Coverage info file. */
char *covFile = "/projects/cc/hg/gs.3/oo.15/bed/coverage/acces_coverage";

/* Disease genes file. */
char *omimFile = "final.omim2gene.txt";
char *extraDiseaseFile = "diseaseGene.txt";

/* Resolved duplications file. */
char *bestDupFile = "bestDup.txt";

/* Centromere/big gap info. */
char *gapFile = "centro.tab";

/* Some colors. */
struct rgbColor rgbRed = {240,0,0};
struct rgbColor rgbRedOrange = {240, 100, 0};
struct rgbColor rgbYellow = {255, 255, 0};

void usage()
/* Explain usage and exit. */
{
errAbort(
  "foldDb - Search database info for making foldout\n"
  "usage:\n"
  "   foldDb chrM chrN ...\n");
}

struct hugeGap 
/* A huge gap to be inserted. */
    {
    struct hugeGap *next;	/* Next in list. */
    int offset;                 /* Offset in ungapped chromosome. */
    int size;                   /* Size of gap. */
    char *type;                 /* Not allocated here. */
    };

int cmpHugeGap(const void *va, const void *vb)
/* Compare to sort biggest offset first. */
{
const struct hugeGap *a = *((struct hugeGap **)va);
const struct hugeGap *b = *((struct hugeGap **)vb);
return a->offset - b->offset;
}

struct chromGaps
/* A list of huge gaps on a chromosome. */
    {
    struct chromGaps *next;
    char *chrom;              /* Name of chromosome, not allocated here. */
    struct hugeGap *gapList;  /* Sorted biggest offset first. */
    };

int gapOffset(struct chromGaps *gaps, int pos)
/* Convert from ungapped to gapped position. */
{
struct hugeGap *gap;
int offset = 0;

if (gaps != NULL)
    {
    for (gap = gaps->gapList; gap != NULL; gap = gap->next)
	{
	if (pos >= gap->offset)
	    {
	    offset += gap->size;
	    pos += gap->size;
	    }
	}
    }
return offset;
}

struct hash *hugeHash;

void setupHugeGaps(struct sqlConnection *conn)
/* Setup things to lookup gaps. */
{
struct lineFile *lf;
char *words[8];
int wordCount;
struct chromGaps *chromList = NULL, *cg;
struct hugeGap *gap;
char *chrom;
char query[512];
struct sqlResult *sr;
char **row;
struct ctgPos ctgPos;
int start, size;
struct hashEl *hel;

hugeHash = newHash(6);
lf = lineFileOpen(gapFile, TRUE);
while ((wordCount = lineFileChop(lf, words)) != 0)
     {
     if (wordCount < 4)
        errAbort("Expecting at least 4 words line %d of %s", lf->lineIx, lf->fileName);
     chrom = words[0];
     if ((cg = hashFindVal(hugeHash, chrom)) == NULL)
         {
	 AllocVar(cg);
	 slAddHead(&chromList, cg);
	 hel = hashAdd(hugeHash, chrom, cg);
	 cg->chrom = hel->name;
	 }
     size = atoi(words[3]);
     sprintf(query, "select * from ctgPos where contig = '%s'", words[2]);
     sr = sqlGetResult(conn, query);
     if ((row = sqlNextRow(sr)) == NULL)
        errAbort("Couldn't find %s from %s in database", words[2], lf->fileName);
     ctgPosStaticLoad(row, &ctgPos);
     if (!sameString(chrom, ctgPos.chrom))
         errAbort("%s is in %s in database and %s in %s", ctgPos.contig, ctgPos.chrom,
	 	chrom, lf->fileName);
     start = ctgPos.chromStart;
     sqlFreeResult(&sr);

     AllocVar(gap);
     slAddHead(&cg->gapList, gap);
     gap->offset = start;
     gap->size = size;
     gap->type = "centromere";
     if (size > 3000000)
         {
	 /* It's more than just a centromere. */
	 gap->size = 3000000;
	 size -= gap->size;
	 start += gap->size;
	 AllocVar(gap);
	 slAddHead(&cg->gapList, gap);
	 gap->offset = start;
	 gap->size = size;
	 gap->type = "other";
	 }
     }
lineFileClose(&lf);
for (cg = chromList; cg != NULL; cg = cg->next)
    slSort(&cg->gapList, cmpHugeGap);
}

void printTab(FILE *f, struct chromGaps *cg, char *chrom, int start, int end, 
	char *track, char *type, int r, int g, int b, char *val)
/* Print line to tab-delimited file. */
{
static double colScale = 1.0/255.0;
double red = r*colScale;
double green = g*colScale;
double blue = b*colScale;

if (!sameString(track, "heterochromatin"))
    {
    int offset = gapOffset(cg, (start+end)/2);
    start += offset;
    end += offset;
    }
fprintf(f, "%s\t%d\t%d\t%s\t%s\t%f,%f,%f\t%s\n",
    chrom, start, end, track, type, red, green, blue, val);
}

void printTabNum(FILE *f, struct chromGaps *cg, char *chrom, int start, int end, 
	char *track, char *type, int r, int g, int b, double num)
/* Print number-valued line to tab-delimited file. */
{
char buf[32];
sprintf(buf, "%f", num);
printTab(f, cg, chrom, start, end, track, type, r, g, b, buf);
}

void getCentrosome(struct chromGaps *cg, FILE *f)
/* Print out centrosome info (and other big gaps). */
{
struct hugeGap *hg;

for (hg = cg->gapList; hg != NULL; hg = hg->next)
    {
    struct rgbColor *col;
    if (sameString("centromere", hg->type))
         col = &rgbRed;
    else
         col = &rgbRedOrange;
    printTab(f, cg, cg->chrom, hg->offset, hg->offset+hg->size,
        "heterochromatin", "box", col->r, col->g, col->b, hg->type);
    }
}

struct resolvedDup
/* A duplication that has been resolved by hand. */
    {
    struct resolvedDup *next;	/* Next in list. */
    char *name;			/* Name - allocated in hash. */
    char *trueChrom;		/* Real chromosome - allocated here. */
    int trueStart;		/* Start of real gene. */
    int trueEnd;		/* End of true gene. */
    boolean used;		/* True when used - can be used only once. */
    };

void makeResolvedDupes(struct hash **retHash, struct resolvedDup **retList)
/* Read in list of dupes that John resolved by hand. */
{
struct hash *hash = newHash(6);
struct hashEl *hel;
struct lineFile *lf = lineFileOpen(bestDupFile, TRUE);
char *words[32], *parts[4];
int wordCount, partCount;
struct resolvedDup *rd, *rdList = NULL;

char *bestDupFile = "bestDup.txt";
while ((wordCount = lineFileChop(lf, words)) > 0)
    {
    lineFileExpectWords(lf, 4, wordCount);
    AllocVar(rd);
    hel = hashAdd(hash, words[0], rd);
    rd->name = hel->name;
    partCount = chopString(words[3], ":-", parts, ArraySize(parts));
    if (partCount != 3)
        errAbort("Misformed chromosome location line %d of %s", lf->lineIx, lf->fileName);
    rd->trueChrom = cloneString(parts[0]);
    rd->trueStart = atoi(parts[1]);
    rd->trueEnd = atoi(parts[2]);
    slAddHead(&rdList, rd);
    }
slReverse(&rdList);
*retHash = hash;
*retList = rdList;
uglyf("Got %d resolved duplications\n", slCount(rdList));
}


boolean isDupe(struct hash *hugoDupHash, struct hash *resolvedDupHash, 
    char *geneName, char *chrom, int start, int end)
/* Return TRUE if it is a dupe that we can't resolve. */
{
struct resolvedDup *rd;
if (!hashLookup(hugoDupHash, geneName))
    return FALSE;
if ((rd = hashFindVal(resolvedDupHash, geneName)) == NULL)
    return TRUE;
uglyf("Found %s in resolvedDupHash\n", geneName);
if (sameString(rd->trueChrom, chrom) && !rd->used)
    {
    uglyf("   %s@%s:%d-%d is true copy\n", geneName, chrom, start, end);
    rd->used = TRUE;
    return FALSE;
    }
else
    return TRUE; 
}


struct omimInfo
/* Info on OMIM gene. */
    {
    struct omimInfo *next;	/* Next in list. */
    char *name;			/* HUGO/OMIM name, allocated in hash. */
    int uses;			/* Number of times used. */
    struct genePred *gpExtra;    /* Extra info if it's on GenieAlt rather than GenieKnown table. */ 
    };


void makeOmimHash(struct sqlConnection *conn,
	struct hash **retHash, struct omimInfo **retList, struct hash **retNameHash)
/* Make hash of all disease genes. */
{
struct hash *hash = newHash(12);
struct hash *nameHash = newHash(12);
struct hashEl *hel;
struct lineFile *lf = lineFileOpen(omimFile, TRUE);
char *words[32];
int wordCount, i;
struct omimInfo *oi, *oiList = NULL;
int oiTotal = 0, oiSingle = 0, oiMultiple = 0, oiSpecial = 0;
char query[256];
struct sqlResult *sr;
char **row;

while ((wordCount = lineFileChop(lf, words)) > 0)
    {
    AllocVar(oi);
    hel = hashAdd(nameHash, words[0], oi);
    oi->name = hel->name;
    for (i=1; i<wordCount; ++i)
        {
	hashAdd(hash, words[i], oi);
	++oi->uses;
	}
    slAddHead(&oiList, oi);
    ++oiTotal;
    if (oi->uses == 1)
       ++oiSingle;
    else if (oi->uses > 1)
       ++oiMultiple;
    }
lineFileClose(&lf);

lf = lineFileOpen(extraDiseaseFile, TRUE);
while ((wordCount = lineFileChop(lf, words)) > 0)
    {
    lineFileExpectWords(lf, 2, wordCount);
    if (!hashLookup(nameHash, words[0]))
        {
	AllocVar(oi);
	hel = hashAdd(nameHash, words[0], oi);
	oi->name = hel->name;
	hashAdd(hash, words[1], oi);
	oi->uses = 1;
	slAddHead(&oiList, oi);
	++oiSpecial;
	sprintf(query, "select * from genieAlt where name = '%s'", words[1]);
	sr = sqlGetResult(conn, query);
	if ((row = sqlNextRow(sr)) != NULL)
	   {
	   oi->gpExtra = genePredLoad(row);
	   printf("Got 'known prediction' %s %s\n", words[0], words[1]);
	   }
	sqlFreeResult(&sr);
	}
    }
lineFileClose(&lf);

slReverse(&oiList);
*retHash = hash;
*retList = oiList;
*retNameHash = nameHash;
uglyf("Got %d omim genes, %d single used, %d multi-use %d special\n", 
	oiTotal, oiSingle, oiMultiple, oiSpecial);
}


struct knownPos
/* Position of a known gene. */
    {
    struct knownPos *next;	/* Next in list. */
    char *name;			/* Name. */
    char *chrom;		/* Chromosome. */
    int chromStart;
    int chromEnd;
    int dupeCount;		/* Number of times duplicated. */
    };


int cmpGenePred(const void *va, const void *vb)
/* Compare to sort based on chromosome, start. */
{
const struct genePred *a = *((struct genePred **)va);
const struct genePred *b = *((struct genePred **)vb);
int dif;
dif = strcmp(a->chrom, b->chrom);
if (dif == 0)
    dif = a->txStart - b->txStart;
return dif;
}


void getKnownGenes(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f, 
	struct hash *hugoDupHash, struct hash *omimHash, struct omimInfo *oiList,
	struct hash *resolvedDupHash, struct hash *omimNameHash, struct hash *knownUniqHash)
/* Get info on known genes. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct genePred *gpList = NULL, *gp;
char geneName[64];
struct omimInfo *oi;
static struct rgbColor diseaseColor = {255, 0, 0};
static struct rgbColor geneColor = { 0, 0, 160};
struct rgbColor *col;
int lastStart = -BIGNUM;
struct knownPos *kp;

printf("  finding known genes\n");

sprintf(query, "select * from genieKnown where chrom = '%s' order by cdsStart", chrom);
sr = sqlGetResult(conn, query);

/* Get list of known genes. */
while ((row = sqlNextRow(sr)) != NULL)
    {
    gp = genePredLoad(row);
    slAddHead(&gpList, gp);
    }
sqlFreeResult(&sr);

/* Add genes that are knowns but in genieAlt. */
for (oi = oiList; oi != NULL; oi = oi->next)
    {
    if (oi->gpExtra && sameString(oi->gpExtra->chrom, chrom))
        {
	slAddHead(&gpList, oi->gpExtra);
	}
    }
slSort(&gpList, cmpGenePred);


/* Now find out name - either from disease gene hash or known table in database. */
for (gp = gpList; gp != NULL; gp = gp->next)
    {
    col = NULL;
    if ((oi = hashFindVal(omimHash, gp->name)) != NULL)
        {
	col = &diseaseColor;
	lastStart = gp->cdsStart;
	strncpy(geneName, oi->name, sizeof(geneName));
	}
    else
	{
	sprintf(query, "select geneName.name from knownInfo,geneName "
		       "where knownInfo.transId = '%s' "
		       "and geneName.id = knownInfo.geneName", gp->name);
	if (sqlQuickQuery(conn, query, geneName, sizeof(geneName)))
	    {
	    if (!sameString(geneName, "n/a") && strlen(geneName) < 8 &&
		    !startsWith("DKFZ", geneName) && !startsWith("KIAA", geneName) && 
		    !hashLookup(omimNameHash, geneName))
		{
		if (gp->cdsStart != lastStart)
		    {
		    col = &geneColor;
		    lastStart = gp->cdsStart;
		    }
		}
	    }
	}
    if (col != NULL)
	{
	if (!isDupe(hugoDupHash, resolvedDupHash, geneName, chrom, gp->cdsStart, gp->cdsEnd))
	    {
	    if (hashLookup(knownUniqHash, geneName))
	        {
		warn("Duplicate %s at %s:%d-%d, skipping all but first\n", 
			geneName, chrom, gp->cdsStart, gp->cdsEnd);
		}
	    else
		{
		hashAdd(knownUniqHash, geneName, NULL);
		printTab(f, cg, chrom, gp->cdsStart, gp->cdsEnd, "hugo", 
			"word", col->r, col->g, col->b, geneName);
		}
	    }
	}
    }
}

#ifdef DECENT_FAKE
void fakeKnownGenes(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get info on known genes. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct hash *mrnaUniq = newHash(0);
struct hash *hugoUniq = newHash(0);
struct psl *pslList = NULL, *psl;
char geneName[64];
int lastStartH = -1, lastEndH = -1;

sprintf(query, "select * from %s_mrna", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    psl = pslLoad(row);
    if (psl->match + psl->repMatch >= psl->qSize/2 && !hashLookup(mrnaUniq, psl->qName))
        {
	hashAdd(mrnaUniq, psl->qName, NULL);
	slAddHead(&pslList, psl);
	}
    else
	pslFree(&psl);
    }
sqlFreeResult(&sr);

slReverse(&pslList);
for (psl = pslList; psl != NULL; psl = psl->next)
    {
    sprintf(query, "select geneName.name from mrna,geneName where mrna.geneName = geneName.id and mrna.acc = '%s'", psl->qName);
    if (sqlQuickQuery(conn, query, geneName, sizeof(geneName)))
        {
	if (!hashLookup(hugoUniq, geneName))
	    {
	    hashAdd(hugoUniq, geneName, NULL);
	    if (!sameString(geneName, "n/a") && strlen(geneName) < 8)
		{
		if (rangeIntersection(lastStartH, lastEndH, psl->tStart, psl->tEnd) <= 0)
		    {
		    printTab(f, cg, chrom, psl->tStart, psl->tEnd, 
			    "hugo", "word", 0, 0, 150, geneName);
		    }
		lastStartH = psl->tStart;
		lastEndH = psl->tEnd;
		}
	    }
	}
    }
freeHash(&mrnaUniq);
freeHash(&hugoUniq);
}
#endif /* DECENT_FAKE */

struct tickPos
/* A tick position */
    {
    struct tickPos *next;
    int start, end;
    char *val;
    };

void getPredGenes(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f, char *table,
  int red, int green, int blue)
/* Get predicted genes. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct genePred *gp;

printf("  Getting %s predicted genes\n", table);
sprintf(query, "select * from %s where chrom = '%s' order by cdsStart", table, chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    gp = genePredLoad(row);
    printTab(f, cg, chrom, gp->cdsStart, gp->cdsEnd, 
	    "genePred", "tick", red, green, blue, ".");
    genePredFree(&gp);
    }
sqlFreeResult(&sr);
}

void getCpgIslands(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get cpgIslands. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct cpgIsland el;

sprintf(query, "select * from cpgIsland where chrom = '%s'", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    cpgIslandStaticLoad(row, &el);
    printTab(f, cg, chrom, el.chromStart, el.chromEnd, 
	    "cpgIsland", "tick", 0, 120, 0, ".");
    }
sqlFreeResult(&sr);
}

void getExofishHits(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get Exofis stuff. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct exoFish el;

sprintf(query, "select * from exoFish where chrom = '%s'", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    exoFishStaticLoad(row, &el);
    printTab(f, cg, chrom, el.chromStart, el.chromEnd, 
	    "tetraodon", "tick", 0, 70, 120, ".");
    }
sqlFreeResult(&sr);
}


void getRnaGenes(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get RNA gene stuff. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct rnaGene el;
int r,g,b;

sprintf(query, "select * from rnaGene where chrom = '%s'", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    rnaGeneStaticLoad(row, &el);
    if (el.isPsuedo)
        {
	r = 255;
	g = 200;
	b = 150;
	}
    else
        {
	r = 120;
	g = 60;
	b = 0;
	}
    printTab(f, cg, chrom, el.chromStart, el.chromEnd, 
	    "rnaGenes", "tick", r, g, b, el.name);
    }
sqlFreeResult(&sr);
}

void getTetHits(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get tetroadon hits. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct wabAli el;

sprintf(query, "select * from waba_tet where chrom = '%s'", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    wabAliStaticLoad(row, &el);
    printTab(f, cg, chrom, el.chromStart, el.chromEnd, 
	    "tetraodon", "tick", 0, 70, 120, ".");
    }
sqlFreeResult(&sr);
}


void getEst3Ticks(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get est 3' ends. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct est3 *el;

sprintf(query, "select * from %s where chrom = '%s'", "est3", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    el = est3Load(row);
    printTab(f, cg, chrom, el->chromStart, el->chromEnd, 
	    "est3", "tick", 0, 0, 0, ".");
    est3Free(&el);
    }
sqlFreeResult(&sr);
}

void getEstIntronTicks(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Get ests with introns. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct psl *psl;
int lastEnd = -BIGNUM;
int sepDistance = 20000;

sprintf(query, "select * from %s_intronEst", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    psl = pslLoad(row);
    if (psl->tStart > lastEnd + sepDistance)
	printTab(f, cg, chrom, psl->tStart, psl->tStart+1, 
		"est3", "tick", 0, 0, 0, ".");
    lastEnd = psl->tEnd;
    pslFree(&psl);
    }
sqlFreeResult(&sr);
}


void getEstTicks(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
{
if (sqlTableExists(conn, "est3"))
   getEst3Ticks(cg, chrom, conn, f);
else
   getEstIntronTicks(cg,chrom,conn, f);
}

void getGc(struct chromGaps *cg, char *chrom, struct sqlConnection *conn, FILE *f)
/* Write wiggle track for GC. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct gcPercent *el;

printf("  getting GC wiggle\n");
sprintf(query, "select * from %s where chrom = '%s'", "gcPercent", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    el = gcPercentLoad(row);
    if (el->gcPpt != 0)
        {
	double scale;
	scale = (el->gcPpt - 300)*0.00300;
	if (scale > 1) scale = 1;
	if (scale < 0) scale = 0;
	printTabNum(f, cg, chrom, el->chromStart, el->chromEnd, 
		"gcPercent", "wiggle", 0, 0, 0, scale);
	}
    gcPercentFree(&el);
    }
sqlFreeResult(&sr);
}

void addSegTotals(int chromStart, int chromEnd, int *countArray, 
	int windowSize, int chromSize)
/* Store the total number of bases in each window between chromStart and chromEnd
 * in the appropriate countArray element. */
{
int startWinIx = chromStart/windowSize;
int endWinIx = (chromEnd-1)/windowSize;
int winBoundary;
int winIx;
int size;

if (chromStart < 0 || chromStart > chromSize)
    {
    warn("Clipping start %d (%d) in addSegTotals\n", chromStart, chromSize);
    return;
    }
if (chromEnd < 0 || chromEnd > chromSize || chromStart >= chromEnd)
    {
    warn("Clipping end %d (%d) in addSegTotals\n", chromStart, chromSize);
    return;
    }
if (startWinIx == endWinIx)
    {
    countArray[startWinIx] += (chromEnd - chromStart);
    return;
    }
winBoundary = (startWinIx+1)*windowSize;
size = winBoundary - chromStart;
assert(size > 0);
assert(size <= windowSize);
countArray[startWinIx] += size;
for (winIx = startWinIx+1; winIx < endWinIx; ++winIx)
    {
    countArray[winIx] += windowSize;
    }
winBoundary = (endWinIx)*windowSize;
size = chromEnd - winBoundary;
assert(size > 0);
assert(size <= windowSize);
countArray[endWinIx] += size;
}

void getRepeatDensity(struct chromGaps *cg, char *chrom, int chromSize, struct sqlConnection *conn, FILE *f,
	char *repClass, int windowSize, double scale, int red, int green, int blue)
/* Put out repeat density info. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct rmskOut rmsk;
struct agpFrag frag;
int winsPerChrom = (chromSize + windowSize-1)/windowSize;
int *winBases;
int *winRepBases;
int i;
int baseCount, repCount;
double density;
double minDen = 1.0, maxDen = 0.0;
int winStart, winEnd;

printf("  Getting %s repeats in %d windows of size %d\n", repClass, winsPerChrom, windowSize);
fflush(stdout);

/* Count up repeats in each window. */
AllocArray(winRepBases, winsPerChrom);
sprintf(query, "select * from %s_rmsk where repClass = '%s'", chrom, repClass);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    rmskOutStaticLoad(row, &rmsk);
    addSegTotals(rmsk.genoStart, rmsk.genoEnd, winRepBases, windowSize, chromSize);
    }
sqlFreeResult(&sr);
/* Count up bases in each window. */
AllocArray(winBases, winsPerChrom);
sprintf(query, "select * from %s_gold", chrom);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    agpFragStaticLoad(row, &frag);
    addSegTotals(frag.chromStart, frag.chromEnd, winBases, windowSize, chromSize);
    }
sqlFreeResult(&sr);

/* Output density. */
for (i=0; i<winsPerChrom; ++i)
    {
    if ((baseCount = winBases[i]) > windowSize/3)
        {
	repCount = winRepBases[i];
	density = (double)repCount/(double)baseCount;
	if (density < minDen) minDen = density;
	if (density > maxDen) maxDen = density;
	density *= scale;
	if (density > 1.0) density = 1.0;
	winStart = i*windowSize;
	winEnd = winStart + windowSize;
	if (winEnd > chromSize) winEnd = chromSize;
	printTabNum(f, cg, chrom, winStart, winEnd, 
		repClass, "wiggle", red, green, blue, density);
	}
    }
printf("   %s minDen %f, maxDen %f\n", chrom, minDen, maxDen);
freeMem(winBases);
freeMem(winRepBases);
}

#define finCov 100

struct rgbColor draftCols[10];


void initDraftCols()
/* Initialize draft colors with a range. */
{
int i;
int dif;
int maxCol = ArraySize(draftCols)-1;
double scale;

for (i=0; i<= maxCol; ++i)
    {
    scale = (double)i/(double)maxCol;
    dif = rgbRedOrange.r - rgbYellow.r;
    draftCols[i].r = rgbYellow.r + round(dif*scale);
    dif = rgbRedOrange.g - rgbYellow.g;
    draftCols[i].g = rgbYellow.g + round(dif*scale);
    dif = rgbRedOrange.b - rgbYellow.b;
    draftCols[i].b = rgbYellow.b + round(dif*scale);
    }
}

void makeBackupCoverageHash(struct sqlConnection *conn, struct hash *hash)
/* Make hash table that describe coverage of each clone just based on
 * whether it's draft/predraft/etc. Only do this if clone not already
 * in hash. */
{
char query[512];
struct clonePos cp;
struct sqlResult *sr;
char **row;
int coverage;
char stage;

sprintf(query, "select * from clonePos");
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    clonePosStaticLoad(row, &cp);
    stage = cp.stage[0];
    if (stage == 'D')
	coverage = 4;
    else if (stage == 'F')
	coverage = finCov;
    else if (stage == 'P')
	coverage = 2;
    else
	errAbort("Unknown stage %c for clone %s", stage, cp.name);
    chopSuffix(cp.name);
    if (!hashLookup(hash, cp.name))
	hashAdd(hash, cp.name, intToPt(coverage));
    }
sqlFreeResult(&sr);
}

void makeMainCoverageHash(char *fileName, struct hash *hash)
/* Scan file for clone coverage. */
{
struct lineFile *lf = lineFileOpen(fileName, TRUE);
char *words[16];
int wordCount;
double d;
int rd;
char *acc;

while ((wordCount = lineFileChop(lf, words)) != 0)
    {
    lineFileExpectWords(lf, wordCount, 4);
    rd = round(atof(words[3]));
    hashAddUnique(hash, words[0], intToPt(rd));
    }
lineFileClose(&lf);
}

struct hash *makeCoverageHash(struct sqlConnection *conn)
/* Get coverage hash, making it if necessary. */
{
static struct hash *hash = NULL;
if (hash == NULL)
    {
    hash = newHash(16);
    makeMainCoverageHash(covFile, hash);
    makeBackupCoverageHash(conn, hash);
    }
return hash;
}


void covInWin(struct hash *covHash, struct chromGaps *cg,
	char *chrom, int start, int end, struct gl *glList, FILE *f)
/* Process coverage in a window of chromosome. */
{
int size = end-start;
UBYTE *cov = needLargeZeroedMem(size+1);  /* A byte for each base in window. */
struct gl *gl;
int s, e;
int coverage;
char cloneName[64];
void *covAsPt;
int i,c;
int startIx, runSize;
UBYTE val, lastVal;

for (gl = glList; gl != NULL; gl = gl->next)
    {
    if ((s = gl->start) < end && (e = gl->end) > start)
        {
	if (s < start) s = start;
	if (e > end) e = end;
	assert(s<e);
	s -= start;
	e -= start;
	strcpy(cloneName, gl->frag);
	chopSuffix(cloneName);
	covAsPt = hashMustFindVal(covHash, cloneName);
	coverage = ptToInt(covAsPt);
	for (i=s; i<e; ++i)
	    {
	    c = coverage + cov[i];
	    if (c > finCov) c = finCov;
	    cov[i] = c;
	    }
	}
    }

/* Add sentinal value to simplify run-length encoding loop. */
cov[size] = finCov+1;
startIx = 0;
lastVal = cov[0];
for (i=1; i<=size; ++i)
    {
    if ((val = cov[i]) != lastVal)
        {
	if (lastVal != 0)
	    {
	    struct rgbColor *col;
	    runSize = i-startIx;
	    if (lastVal == finCov)
	        col = &rgbRed;
	    else if (lastVal >= ArraySize(draftCols))
	        col = &rgbRedOrange;
	    else
	        col = &draftCols[lastVal-1];
	    s = startIx + start;
	    e = i + start;
	    printTabNum(f, cg, chrom, s, e, 
		    "coverage", "box", col->r, col->g, col->b, lastVal);
	    }
	startIx = i;
	lastVal = val;
	}
    }
freeMem(cov);
}

void getCoverage(struct chromGaps *cg, char *chrom, int chromSize, struct sqlConnection *conn, FILE *f)
/* Make coverage track. */
{
struct hash *covHash = makeCoverageHash(conn);
char query[512];
struct gl *glList = NULL, *gl;
struct sqlResult *sr;
char **row;
int coverage;
int start, end;
int winSize = 1024*1024*64;

printf("  getting coverage\n");
initDraftCols();
sprintf(query, "select * from %s_gl", chrom);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    gl = glLoad(row);
    slAddHead(&glList, gl);
    }
sqlFreeResult(&sr);
slReverse(&glList);

for (start = 0; start < chromSize; start = end)
    {
    end = start + winSize;
    if (end > chromSize)
        end = chromSize;
    covInWin(covHash, cg, chrom, start, end, glList, f);
    }
glFreeList(&glList);
}

void getEnsTicks(struct chromGaps *cg, char *chrom, int chromSize, 
	FILE *f, char *tabName)
/* Get Ensembl gene ticks from tab separated file. */
{
struct lineFile *lf = lineFileOpen(tabName, TRUE);
char *words[8];
int wordCount;
int start;

printf("  Getting Ensembl gene ticks\n");
while ((wordCount = lineFileChop(lf, words)) != 0)
    {
    lineFileExpectWords(lf, 4, wordCount);
    if (sameString(words[0], chrom))
        {
	start = atoi(words[1]);
	printTab(f, cg, chrom, start, start+1, 
		"genePred", "tick", 150, 0, 0, ".");
	}
    }
lineFileClose(&lf);
}

double nRatio(struct sqlConnection *conn, 
	char *chrom, int chromStart, int chromEnd)
/* Return percentage of N's in range. */
{
char query[256];
struct sqlResult *sr;
char **row;
struct agpGap gap;
int baseTotal = chromEnd - chromStart;
int nCount = 0, s, e, size;

sprintf(query, "select * from %s_gap where chromStart < %d and chromEnd > %d",
	chrom, chromEnd, chromStart);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    agpGapStaticLoad(row, &gap);
    s = gap.chromStart;
    if (s < chromStart) s = chromStart;
    e = gap.chromEnd;
    if (e > chromEnd) e = chromEnd;
    size = e - s;
    nCount += size;
    }
sqlFreeResult(&sr);
return (double)nCount/(double)baseTotal;
}

int countSnps(struct sqlConnection *conn, 
	char *chrom, int chromStart, int chromEnd)
/* Count the number of SNPs in window. */
{
char query[256];

sprintf(query, 
    "select count(*) from snpTsc where chrom = '%s' and chromStart < %d and chromEnd > %d",
    chrom, chromEnd, chromStart);
return sqlQuickNum(conn, query);
}

void slowGetSnpDensity(struct chromGaps *cg, char *chrom, int chromSize, struct sqlConnection *conn, FILE *f)
/* Make SNP density track. */
{
int windowSize = 50000;
int s, e;
int size;
double ratio, totalRatio = 0;
double maxCount = 0, minCount=windowSize*2, totalCount = 0;

for (s = 0; s < chromSize; s += 50000)
    {
    e = s + windowSize;
    if (e > chromSize) e = chromSize;
    size = e - s;
    if (size > windowSize/2)
        {
	ratio = 1.0 - nRatio(conn, chrom, s, e);
	uglyf("%s:%d-%d ratio %f\n", chrom, s, e, ratio);
	if (ratio >= 0.5)
	    {
	    int rawCount = countSnps(conn, chrom, s, e);
	    double count = rawCount/ratio;
	    uglyf(" rawCount %d, count %f\n", rawCount, count);
	    if (count < minCount)
	        minCount = count;
	    if (count > maxCount)
	        maxCount = count;
	    totalRatio += ratio;
	    totalCount += count;
	    }
	}
    }
printf("%s SNP density in %d windows is between %f and %f, average %f\n",
	chrom, windowSize, minCount, maxCount, totalCount/totalRatio);
}

void getSnpDensity(struct chromGaps *cg, char *chrom, int chromSize, struct sqlConnection *conn, FILE *f)
/* Put out SNP density info. */
{
int windowSize = 50000;
double scale = 500.0;
struct sqlResult *sr;
char **row;
char query[256];
struct snp snp;
struct agpFrag frag;
int winsPerChrom = (chromSize + windowSize-1)/windowSize;
int *winBases;
int *winRepBases;
int i;
int baseCount, repCount;
double density;
double minDen = 1.0, maxDen = 0.0;
int winStart, winEnd;

printf("  Getting SNPs in %d windows of size %d\n", winsPerChrom, windowSize);
fflush(stdout);

/* Count up SNPs in each window. */
AllocArray(winRepBases, winsPerChrom);
sprintf(query, "select * from snpTsc where chrom = '%s'", chrom);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    snpStaticLoad(row, &snp);
    addSegTotals(snp.chromStart, snp.chromEnd, winRepBases, windowSize, chromSize);
    }
sqlFreeResult(&sr);

/* Count up bases in each window. */
AllocArray(winBases, winsPerChrom);
sprintf(query, "select * from %s_gold", chrom);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    agpFragStaticLoad(row, &frag);
    addSegTotals(frag.chromStart, frag.chromEnd, winBases, windowSize, chromSize);
    }
sqlFreeResult(&sr);

/* Output density. */
for (i=0; i<winsPerChrom; ++i)
    {
    if ((baseCount = winBases[i]) > windowSize/2)
        {
	repCount = winRepBases[i];
	density = (double)repCount/(double)baseCount;
	if (density < minDen) minDen = density;
	if (density > maxDen) maxDen = density;
	density *= scale;
	if (density > 1.0) density = 1.0;
	winStart = i*windowSize;
	winEnd = winStart + windowSize;
	if (winEnd > chromSize) winEnd = chromSize;
	printTabNum(f, cg, chrom, winStart, winEnd, 
		"SNP", "wiggle", 128, 0, 128, density);
	}
    }
printf("   %s minDen %f, maxDen %f\n", chrom, minDen, maxDen);
freeMem(winBases);
freeMem(winRepBases);
}

void getBands(struct chromGaps *cg, char *chrom, int chromSize, struct sqlConnection *conn, FILE *f)
/* Get chromosome bands stuff. */
{
struct sqlResult *sr;
char **row;
char query[256];
struct cytoBand el;
int r,g,b;
char *stain;

printf("  getting bands\n");
sprintf(query, "select * from cytoBand where chrom = '%s'", chrom);
sr = sqlGetResult(conn, query);

while ((row = sqlNextRow(sr)) != NULL)
    {
    cytoBandStaticLoad(row, &el);
    stain = el.gieStain;
    if (startsWith("gneg", stain))
        {
	r = g = b = 220;
	}
    else if (startsWith ("gpos", stain))
        {
	int percentage = 100;
	stain += 4;
	if (isdigit(stain[0]))
	    percentage = atoi(stain);
	r = g = b = round((100.0-percentage)*200);
	}
    else if (startsWith("gvar", stain))
        {
	r = 180;
	g = 180;
	b = 120;
	}
    else
        {
	r = 150;
	g = 50;
	b = 50;
	}
    printTab(f, cg, chrom, el.chromStart, el.chromEnd, 
	    "band", "box", r, g, b, el.name);
    }
sqlFreeResult(&sr);
}


void oneChrom(char *chrom, struct sqlConnection *conn, FILE *f,
	struct hash *hugoDupHash, struct hash *omimHash, struct omimInfo *oiList,
	struct hash *resolvedDupHash, struct hash *omimNameHash, 
	struct hash *knownUniqHash)
/* Get info for one chromosome.  */
{
int chromSize = hChromSize(chrom);
struct chromGaps *cg = hashFindVal(hugeHash, chrom);
printf("Processing %s\n", chrom);
if (cg != NULL)
    getCentrosome(cg, f);

getBands(cg, chrom, chromSize, conn, f);
getCoverage(cg, chrom, chromSize, conn, f);
getGc(cg, chrom, conn, f);
getRepeatDensity(cg, chrom, chromSize, conn, f, "SINE", 100000, 1.0/0.66, 255, 0, 0);
getRepeatDensity(cg, chrom, chromSize, conn, f, "LINE", 100000, 1.0/0.66, 0, 0, 255);
getSnpDensity(cg, chrom, chromSize, conn, f);
getRnaGenes(cg, chrom, conn, f);
getCpgIslands(cg, chrom, conn, f);
//getTetHits(cg, chrom, conn, f);
getExofishHits(cg,chrom,conn,f);
getEstTicks(cg, chrom, conn, f);
#ifdef SOON
getPredGenes(cg, chrom, conn, f, "ensGene", 150, 0, 0);
#endif SOON
getPredGenes(cg, chrom, conn, f, "genieAlt", 150, 0, 0);
getEnsTicks(cg, chrom, chromSize, f, "ens/chrom.bed");
getPredGenes(cg, chrom, conn, f, "genieKnown", 0, 0, 150);
getKnownGenes(cg, chrom, conn, f, hugoDupHash, omimHash, oiList,
	resolvedDupHash, omimNameHash, knownUniqHash);
}

void getHugoDupes(struct sqlConnection *conn, struct hash *hugoDupHash)
/* Read in all hugo genes and save the duplicated ones in hugoDupHash */
{
struct hash *hash = newHash(0);
struct sqlResult *sr;
char **row;
char *name;
struct slName *txIdList = NULL, *txId;
char query[256];
char geneName[256];

printf("Looking for duplicated HUGO genes:\n");
sr = sqlGetResult(conn, "select name from genieKnown");
while ((row = sqlNextRow(sr)) != NULL)
    {
    txId = newSlName(row[0]);
    slAddHead(&txIdList, txId);
    }
sqlFreeResult(&sr);


for (txId = txIdList; txId != NULL; txId = txId->next)
    {
    sprintf(query, "select geneName.name from knownInfo,geneName "
                   "where knownInfo.transId = '%s' "
		   "and geneName.id = knownInfo.geneName", txId->name);
    if (sqlQuickQuery(conn, query, geneName, sizeof(geneName)))
    if (hashLookup(hash, geneName) && !hashLookup(hugoDupHash, geneName))
	{
	printf("    %s\n", geneName);
        hashStore(hugoDupHash, geneName);
	}
    else
        hashAdd(hash, geneName, NULL);
    }
freeHash(&hash);
}

void foldDb(int chromCount, char *chromNames[])
/* foldDb - Search database info for making foldout. */
{
int i;
char *database = "hg4";
struct sqlConnection *conn = sqlConnect(database);
struct hash *hugoDupHash = newHash(0);
struct hash *knownUniqHash = newHash(0);
struct hash *omimHash = NULL, *omimNameHash = NULL;
struct hash *resolvedDupHash = NULL;
struct resolvedDup *rdList = NULL;
struct omimInfo *oiList = NULL;

hSetDb(database);
setupHugeGaps(conn);
getHugoDupes(conn, hugoDupHash);
makeOmimHash(conn, &omimHash, &oiList, &omimNameHash);
makeResolvedDupes(&resolvedDupHash, &rdList);
for (i=0; i<chromCount; ++i)
    {
    char fileName[512];
    FILE *f;
    sprintf(fileName, "%s.tab", chromNames[i]);
    f = mustOpen(fileName, "w");
    oneChrom(chromNames[i], conn, f, hugoDupHash, omimHash, oiList,
    	resolvedDupHash, omimNameHash, knownUniqHash);
    fclose(f);
    }
sqlDisconnect(&conn);
}

int main(int argc, char *argv[])
/* Process command line. */
{
if (argc < 2)
    usage();
foldDb(argc-1, argv+1);
return 0;
}
